import { Observable } from "@microsoft/fast-element";
class DerivedValueEvaluator {
    constructor(value) {
        this.value = value;
        this.notifier = Observable.getNotifier(this);
        this.dependencies = new Set();
        this.binding = Observable.binding(value, this);
        this.binding.setMode(false);
    }
    static getOrCreate(value) {
        let v = DerivedValueEvaluator.cache.get(value);
        if (v) {
            return v;
        }
        v = new DerivedValueEvaluator(value);
        DerivedValueEvaluator.cache.set(value, v);
        return v;
    }
    evaluate(node, tokenContext) {
        const resolve = (token) => {
            this.dependencies.add(token);
            if (tokenContext === token) {
                if (node.parent) {
                    return node.parent.getTokenValue(token);
                }
                throw new Error("DesignTokenNode has encountered a circular token reference. Avoid this by setting the token value for an ancestor node.");
            }
            else {
                return node.getTokenValue(token);
            }
        };
        return this.binding.observe(resolve);
    }
    handleChange() {
        this.notifier.notify(undefined);
    }
}
DerivedValueEvaluator.cache = new WeakMap();
class DerivedValue {
    constructor(token, evaluator, node, subscriber) {
        this.token = token;
        this.evaluator = evaluator;
        this.node = node;
        this.subscriber = subscriber;
        this.value = evaluator.evaluate(node, token);
        if (this.subscriber) {
            Observable.getNotifier(this.evaluator).subscribe(this.subscriber);
        }
    }
    dispose() {
        if (this.subscriber) {
            Observable.getNotifier(this.evaluator).unsubscribe(this.subscriber);
        }
    }
    update() {
        this.value = this.evaluator.evaluate(this.node, this.token);
        return this;
    }
}
/**
 * @internal
 */
export class DesignTokenChangeRecordImpl {
    constructor(target, type, token, value) {
        this.target = target;
        this.type = type;
        this.token = token;
        this.value = value;
    }
    notify() {
        Observable.getNotifier(this.token).notify(this);
    }
}
/**
 * @public
 */
export class DesignTokenNode {
    constructor() {
        this._parent = null;
        this._children = new Set();
        this._values = new Map();
        this._derived = new Map();
        this.dependencyGraph = new Map();
    }
    /**
     * Determines if a value is a {@link DerivedDesignTokenValue}
     * @param value - The value to test
     */
    static isDerivedTokenValue(value) {
        return typeof value === "function";
    }
    /**
     * Determines if a token has a derived value for a node.
     */
    static isDerivedFor(node, token) {
        return node._derived.has(token);
    }
    /**
     * Collects token/value pairs for all derived token / values set on upstream nodes.
     */
    static collectDerivedContext(node) {
        const collected = new Map();
        // Exit early if  there is no parent
        if (node.parent === null) {
            return collected;
        }
        let ignored = DesignTokenNode.getAssignedTokensForNode(node);
        let current = node.parent;
        do {
            const assigned = DesignTokenNode.getAssignedTokensForNode(current);
            for (let i = 0, l = assigned.length; i < l; i++) {
                const token = assigned[i];
                if (!ignored.includes(token) &&
                    DesignTokenNode.isDerivedFor(current, token)) {
                    collected.set(token, current._derived.get(token));
                }
            }
            ignored = Array.from(new Set(ignored.concat(assigned)));
            current = current.parent;
        } while (current !== null);
        return collected;
    }
    /**
     * Resolves the local value for a token if it is assigned, otherwise returns undefined.
     */
    static getLocalTokenValue(node, token) {
        return !DesignTokenNode.isAssigned(node, token)
            ? undefined
            : DesignTokenNode.isDerivedFor(node, token)
                ? node._derived.get(token).value
                : node._values.get(token);
    }
    static getOrCreateDependencyGraph(node, token) {
        let dependents = node.dependencyGraph.get(token);
        if (dependents) {
            return dependents;
        }
        dependents = new Set();
        node.dependencyGraph.set(token, dependents);
        return dependents;
    }
    /**
     * Emit all queued notifications
     */
    static notify() {
        const notifications = this._notifications;
        this._notifications = [];
        for (const record of notifications) {
            record.notify();
        }
    }
    static queueNotification(...records) {
        this._notifications.push(...records);
    }
    /**
     * Retrieves all tokens assigned directly to a node.
     * @param node - the node to retrieve assigned tokens for
     * @returns
     */
    static getAssignedTokensForNode(node) {
        return Array.from(node._values.keys());
    }
    /**
     * Retrieves all tokens assigned to the node and ancestor nodes.
     * @param node - the node to compose assigned tokens for
     */
    static composeAssignedTokensForNode(node) {
        const tokens = new Set(DesignTokenNode.getAssignedTokensForNode(node));
        let current = node.parent;
        while (current !== null) {
            const assignedTokens = DesignTokenNode.getAssignedTokensForNode(current);
            for (const token of assignedTokens) {
                tokens.add(token);
            }
            current = current.parent;
        }
        return Array.from(tokens);
    }
    /**
     * Tests if a token is assigned directly to a node
     * @param node - The node to test
     * @param token  - The token to test
     * @returns
     */
    static isAssigned(node, token) {
        return node._values.has(token);
    }
    /**
     * The parent node
     */
    get parent() {
        return this._parent;
    }
    get children() {
        return Array.from(this._children);
    }
    /**
     * Appends a child to the node, notifying for any tokens set for the node's context.
     */
    appendChild(child) {
        var _a, _b;
        let prevContext = null;
        // If this node is already attached, get it's context so change record
        // types can be determined
        if (child.parent !== null) {
            prevContext = DesignTokenNode.composeAssignedTokensForNode(child.parent);
            child.parent._children.delete(child);
        }
        const context = DesignTokenNode.composeAssignedTokensForNode(this);
        const derivedContext = DesignTokenNode.collectDerivedContext(this);
        child._parent = this;
        this._children.add(child);
        for (const token of context) {
            let type = 0 /* DesignTokenMutationType.add */;
            if (prevContext !== null) {
                const prevContextIndex = prevContext.indexOf(token);
                if (prevContextIndex !== -1) {
                    type = 1 /* DesignTokenMutationType.change */;
                    prevContext.splice(prevContextIndex, 1);
                }
            }
            child.dispatch(new DesignTokenChangeRecordImpl(this, type, token, (_a = derivedContext.get(token)) === null || _a === void 0 ? void 0 : _a.evaluator.value));
        }
        if (prevContext !== null && prevContext.length > 0) {
            for (const token of prevContext) {
                child.dispatch(new DesignTokenChangeRecordImpl(this, 2 /* DesignTokenMutationType.delete */, token, (_b = derivedContext.get(token)) === null || _b === void 0 ? void 0 : _b.evaluator.value));
            }
        }
        DesignTokenNode.notify();
    }
    /**
     * Appends a child to the node, notifying for any tokens set for the node's context.
     */
    removeChild(child) {
        if (child.parent === this) {
            const context = DesignTokenNode.composeAssignedTokensForNode(this);
            child._parent = null;
            this._children.delete(child);
            for (const token of context) {
                child.dispatch(new DesignTokenChangeRecordImpl(this, 2 /* DesignTokenMutationType.delete */, token));
            }
            DesignTokenNode.notify();
        }
    }
    /**
     * Dispose of the node, removing parent/child relationships and
     * unsubscribing all observable binding subscribers. Does not emit
     * notifications.
     */
    dispose() {
        if (this.parent) {
            this.parent._children.delete(this);
            this._parent = null;
        }
        for (const [, derived] of this._derived) {
            derived.dispose();
        }
    }
    /**
     * Sets a token to a value
     */
    setTokenValue(token, value) {
        const changeType = DesignTokenNode.isAssigned(this, token) ||
            DesignTokenNode.isDerivedFor(this, token)
            ? 1 /* DesignTokenMutationType.change */
            : 0 /* DesignTokenMutationType.add */;
        const prev = DesignTokenNode.getLocalTokenValue(this, token);
        this._values.set(token, value);
        if (DesignTokenNode.isDerivedFor(this, token)) {
            this.tearDownDerivedTokenValue(token);
        }
        const isDerived = DesignTokenNode.isDerivedTokenValue(value);
        const derivedContext = DesignTokenNode.collectDerivedContext(this);
        let result;
        if (isDerived) {
            const evaluator = this.setupDerivedTokenValue(token, value, true);
            result = evaluator.value;
        }
        else {
            result = value;
        }
        if (prev !== result) {
            DesignTokenNode.queueNotification(new DesignTokenChangeRecordImpl(this, changeType, token, value));
        }
        this.dispatch(new DesignTokenChangeRecordImpl(this, changeType, token, value));
        derivedContext.forEach((derivedValue, token) => {
            // Skip over any derived values already established locally, because
            // those will get updated via this.notifyDerived and this.notifyStatic
            if (!DesignTokenNode.isDerivedFor(this, token)) {
                const prev = DesignTokenNode.getLocalTokenValue(this, token);
                derivedValue = this.setupDerivedTokenValue(token, derivedValue.evaluator.value);
                const result = derivedValue.value;
                if (prev !== result) {
                    DesignTokenNode.queueNotification(new DesignTokenChangeRecordImpl(this, 1 /* DesignTokenMutationType.change */, token, derivedValue.evaluator.value));
                }
                this.dispatch(new DesignTokenChangeRecordImpl(this, 0 /* DesignTokenMutationType.add */, token, derivedValue.evaluator.value));
            }
        });
        DesignTokenNode.notify();
    }
    /**
     * Returns the resolve value for a token
     */
    getTokenValue(token) {
        /* eslint-disable-next-line */
        let node = this;
        let value;
        while (node !== null) {
            if (DesignTokenNode.isDerivedFor(node, token)) {
                value = node._derived.get(token).value;
                break;
            }
            if (DesignTokenNode.isAssigned(node, token)) {
                value = node._values.get(token);
                break;
            }
            node = node._parent;
        }
        if (value !== undefined) {
            return value;
        }
        else {
            throw new Error(`No value set for token ${token} in node tree.`);
        }
    }
    /**
     * Deletes the token value for a node
     */
    deleteTokenValue(token) {
        if (DesignTokenNode.isAssigned(this, token)) {
            const prev = DesignTokenNode.getLocalTokenValue(this, token);
            this._values.delete(token);
            this.tearDownDerivedTokenValue(token);
            let newValue;
            try {
                newValue = this.getTokenValue(token);
            }
            catch (e) {
                newValue = undefined;
            }
            DesignTokenNode.queueNotification(new DesignTokenChangeRecordImpl(this, 2 /* DesignTokenMutationType.delete */, token));
            if (prev !== newValue) {
                this.dispatch(new DesignTokenChangeRecordImpl(this, 2 /* DesignTokenMutationType.delete */, token));
            }
            DesignTokenNode.notify();
        }
    }
    /**
     * Notifies that a token has been mutated
     */
    dispatch(record) {
        var _a, _b, _c;
        if (this !== record.target) {
            const { token } = record;
            // If the node is assigned the token being dispatched and the assigned value does not depend on the token
            // (circular token reference) then terminate the dispatch.
            const isAssigned = DesignTokenNode.isAssigned(this, token);
            const containsCircularForToken = isAssigned && ((_a = this._derived.get(token)) === null || _a === void 0 ? void 0 : _a.evaluator.dependencies.has(token));
            if (isAssigned && !containsCircularForToken) {
                return;
            }
            // Delete token evaluations if the token is not assigned explicitly but is derived for the node and
            // the record is a delete type.
            if (record.type === 2 /* DesignTokenMutationType.delete */ &&
                !isAssigned &&
                DesignTokenNode.isDerivedFor(this, token)) {
                this.tearDownDerivedTokenValue(token);
                DesignTokenNode.queueNotification(new DesignTokenChangeRecordImpl(this, 2 /* DesignTokenMutationType.delete */, token));
            }
            if (containsCircularForToken) {
                record = new DesignTokenChangeRecordImpl(this, 1 /* DesignTokenMutationType.change */, token, (_b = this._derived.get(token)) === null || _b === void 0 ? void 0 : _b.evaluator.value);
            }
            const { value } = record;
            if (value && DesignTokenNode.isDerivedTokenValue(value)) {
                const dependencies = DerivedValueEvaluator.getOrCreate(value)
                    .dependencies;
                // If this is not the originator, check to see if this node
                // has any dependencies of the token value. If so, we need to evaluate for this node
                let evaluate = false;
                for (const dependency of dependencies) {
                    if (DesignTokenNode.isAssigned(this, dependency)) {
                        evaluate = true;
                        break;
                    }
                }
                if (evaluate) {
                    const prev = (_c = this._derived.get(token)) === null || _c === void 0 ? void 0 : _c.value;
                    const derivedValue = this.setupDerivedTokenValue(token, value);
                    if (prev !== derivedValue.value) {
                        const type = prev === undefined
                            ? 0 /* DesignTokenMutationType.add */
                            : 1 /* DesignTokenMutationType.change */;
                        const notification = new DesignTokenChangeRecordImpl(this, type, token, derivedValue.evaluator.value);
                        DesignTokenNode.queueNotification(notification);
                        record = notification;
                    }
                }
            }
        }
        this.collectLocalChangeRecords(record).forEach(_record => {
            DesignTokenNode.queueNotification(_record);
            this.dispatch(_record);
        });
        this.notifyChildren(record);
    }
    /**
     * Generate change-records for local dependencies of a change record
     */
    collectLocalChangeRecords(record) {
        const collected = new Map();
        for (const dependent of DesignTokenNode.getOrCreateDependencyGraph(this, record.token)) {
            if (dependent.value !== dependent.update().value) {
                collected.set(dependent.token, new DesignTokenChangeRecordImpl(this, 1 /* DesignTokenMutationType.change */, dependent.token, dependent.evaluator.value));
            }
        }
        return collected;
    }
    /**
     *
     * Notify children of changes to the node
     */
    notifyChildren(...records) {
        if (this.children.length) {
            for (let i = 0, l = this.children.length; i < l; i++) {
                for (let j = 0; j < records.length; j++) {
                    this.children[i].dispatch(records[j]);
                }
            }
        }
    }
    tearDownDerivedTokenValue(token) {
        if (DesignTokenNode.isDerivedFor(this, token)) {
            const value = this._derived.get(token);
            value.dispose();
            this._derived.delete(token);
            value.evaluator.dependencies.forEach(dependency => {
                DesignTokenNode.getOrCreateDependencyGraph(this, dependency).delete(value);
            });
        }
    }
    setupDerivedTokenValue(token, value, subscribeNode = false) {
        const deriver = new DerivedValue(token, DerivedValueEvaluator.getOrCreate(value), this, subscribeNode
            ? {
                handleChange: () => {
                    if (deriver.value !== deriver.update().value) {
                        const record = new DesignTokenChangeRecordImpl(this, 1 /* DesignTokenMutationType.change */, deriver.token, deriver.evaluator.value);
                        DesignTokenNode.queueNotification(record);
                        this.dispatch(record);
                        DesignTokenNode.notify();
                    }
                },
            }
            : undefined);
        this._derived.set(token, deriver);
        deriver.evaluator.dependencies.forEach(dependency => {
            if (dependency !== token) {
                DesignTokenNode.getOrCreateDependencyGraph(this, dependency).add(deriver);
            }
        });
        return deriver;
    }
}
DesignTokenNode._notifications = [];
